'''
utils.py
part of the wordfish python tools

'''

from urllib2 import Request, urlopen, HTTPError
from glob import glob
import errno
import tarfile
import zipfile
import urllib2
import json
import shutil
import os
import re
import __init__
import pickle

def wordfish_home():
    return os.environ["WORDFISH_HOME"]

def save_pretty_json(dict_obj,output_file):
    with open(output_file, 'w') as outfile:
        json.dump(dict_obj,outfile)

def get_installdir():
    return os.path.dirname(os.path.abspath(__init__.__file__))

def save_pkl(save_obj,pickle_file):
    pickle.dump(save_obj,open(pickle_file,"wb"))

def mkdir(dirname):
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    return os.path.abspath(dirname)

def find_subdirectories(basepath):
    '''find_subdirectories
    Return directories (and sub) starting from a base
    '''
    directories = []
    for root, dirnames, filenames in os.walk(basepath):
        new_directories = [d for d in dirnames if d not in directories]
        directories = directories + new_directories
    return directories
    
'''
Make a directory if it doesn't exist

'''

def make_directory(dirname):
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    return dirname

'''
Return directories at one level specified by user
(not recursive)

'''
def find_directories(root,fullpath=True):
    directories = []
    for item in os.listdir(root):
        # Don't include hidden directories
        if not re.match("^[.]",item):
            if os.path.isdir(os.path.join(root, item)):
                if fullpath:
                    directories.append(os.path.abspath(os.path.join(root, item)))
                else:
                    directories.append(item)
    return directories

"""
remove unicode keys and values from dict, encoding in utf8

"""
def remove_unicode_dict(input_dict,encoding="utf-8"):
    output_dict = dict()
    for key,value in input_dict.iteritems():
        if isinstance(input_dict[key],list):
            output_new = [x.encode(encoding) for x in input_dict[key]]
        elif isinstance(input_dict[key],int):
            output_new = input_dict[key]
        elif isinstance(input_dict[key],float):
            output_new = input_dict[key]
        elif isinstance(input_dict[key],dict):
            output_new = remove_unicode_dict(input_dict[key])
        else:
            output_new = input_dict[key].encode(encoding)
        output_dict[key.encode(encoding)] = output_new
    return output_dict

"""
Copy an entire directory recursively

"""
 
def copy_directory(src, dest):
    try:
        shutil.copytree(src, dest)
    except OSError as e:
        # If the error was caused because the source wasn't a directory
        if e.errno == errno.ENOTDIR:
            shutil.copy(src, dest)
        else:
            print('Directory not copied. Error: %s' % e)

"""
get_template: read in and return a template file

"""
def get_template(template_file):
    filey = open(template_file,"rb")
    template = "".join(filey.readlines())
    filey.close()
    return template

"""
make a substitution for a template_tag in a template
"""
def sub_template(template,template_tag,substitution):
    template = template.replace(template_tag,substitution)
    return template

def save_template(output_file,html_snippet):
    filey = open(output_file,"w")
    filey.writelines(html_snippet)
    filey.close()

def untar(tar_file,destination="."):
    tar = tarfile.open(tar_file)
    tar.extractall(path=destination)
    tar.close()
    return tar

"""
Check type
"""
def is_type(var,types=[int,float,list]):
    for x in range(len(types)):
        if isinstance(var,types[x]):
            return True
    return False

"""
Ensure utf-8
"""
def clean_fields(mydict):
    newdict = dict()
    for field,value in mydict.iteritems():
        cleanfield = field.encode("utf-8")
        if isinstance(value,float):
            newdict[cleanfield] = value
        if isinstance(value,int):
            newdict[cleanfield] = value
        if isinstance(value,list):
            newlist = []
            for x in value:
                if not is_type(x):
                    newlist.append(x.encode("utf-8"))
                else:
                    newlist.append(x)
            newdict[cleanfield] = newlist
        else:
            newdict[cleanfield] = value.encode("utf-8")
    return newdict

# JOBS #####################################################

def add_lines(script,lines_to_add):
    if isinstance(lines_to_add,str):
        lines_to_add = [lines_to_add]
    lines = []
    if os.path.exists(script):
        filey = open(script,"rb")
        lines = [x.strip("\n") for x in filey.readlines()]
        filey.close()
    for new_line in lines_to_add:
        lines.append(new_line)
    filey = open(script,"wb")
    filey.writelines("\n".join(lines))
    filey.close()

# INTERNET ################################################
def get_url(url):
    request = Request(url)
    response = urlopen(request)
    return response.read()

def get_json(url):
    '''Return general json'''
    print url
    json_single = get_url(url)
    return json.loads(json_single.decode("utf-8"))

def read_json(json_file):
    return json.load(open(json_file,"rb"))

def has_internet_connectivity():
    """has_internet_connectivity
    Checks for internet connectivity by way of trying to
    retrieve google IP address. Returns True/False
    """
    try:
        response=urllib2.urlopen('http://www.google.com',timeout=1)
        return True
    except urllib2.URLError as err: pass
    return False

def make_zip(path,out_zip):
    zipper = zipfile.ZipFile(out_zip, 'w')
    for root, dirs, files in os.walk(path):
        for file in files:
            zipper.write(os.path.join(root, file))
    zipper.close()
    return out_zip



'''
database.py

part of the wordfish python package: extracting relationships of terms from corpus


'''

import os

def import_files(files,table_name):
    if isinstance(files,str): 
        files = [files]
    for import_file in files:
        if os.path.exists(import_file):
            os.system('deepdive sql "COPY %s FROM STDIN CSV" <%s' %(table_name,import_file))

def create_all_tables():
    create_mentions_table()
    create_sentences(table)
    create_features_table()
    create_inference_table()


def create_table():
    print "WRITEME"


def make_mentions_table():
    print "WRITEME"


def make_sentences_table():
    print "WRITEME"




'''
corpus.py

part of the wordfish python package: extracting relationships of terms from corpus
this set of functions works with different plugins (in plugins folder) to produce input corpus
for parsing.

'''

from wordfish.utils import find_directories, save_pretty_json
from wordfish.nlp import remove_nonenglish_chars
from textblob import TextBlob
from glob import glob
import os

            
def save_sentences(articles,output_dir=".",prefix=""):
    '''save_sentences: 
    parses a dictionary with articles and sentences into output sentence files
    Parameters
    ==========
    articles: list or dict
        a list or dictionary of articles to parse and save to sentences.txt files.
        If you have a meaningful unique ID, provide a dictionary with the UID as key
        If not, provide a list and an integer will be generated instead.
    output_dir: path
        full path that corresponds to the plugins specific directory under "corpus"
        eg $WORK/corpus/neurosynth
    '''
    if isinstance(articles,dict):
        for uid, meta in articles.iteritems():
            meta["text"] = remove_nonenglish_chars(meta["text"])
            save_sentences_single(uid,meta["text"],output_dir,prefix)
            # If more than one entry is provided, we also have meta data
            if len(meta)>1:
                save_meta(uid,meta,output_dir)

    elif isinstance(articles,list):
        for uid in range(len(articles)):
            articles["uid"] = remove_nonenglish_chars(articles["uid"])
            save_sentences_single(uid,articles[uid],output_dir,prefix)

def save_sentences_single(uid,text,output_dir,remove_non_english_chars=True,prefix=""):
    '''save_sentence_single
    Write text to file 
    Parameters
    ==========
    uid: int or string
        a unique ID for the article
    text: str
        raw text of the article    
    output_dir: path
        full path to a plugins corpus directory
    '''
    if prefix != "":
        prefix = "%s_" %(prefix)
    output_file = "%s/%s%s_sentences.txt" %(output_dir,uid,prefix)
    filey = open(output_file,"wb")
    filey.write("%s\n" %(text))
    filey.close()


def save_meta(uid,meta,output_dir,prefix=""):
    '''save_meta
 
    Parameters
    ==========
    uid: int or string
        a unique ID for the article
    meta: dict
        dictionary with meta info, and labels
    output_dir: path
        full path to a plugins corpus directory
    '''
    if prefix != "":
        prefix = "%s_" %(prefix)
    output_file = "%s/%s_%smeta.txt" %(output_dir,uid,prefix)
    tmp = save_pretty_json(meta,output_file)
    return tmp

def get_meta(base_dir):
    corpus_folders = find_directories("%s/corpus" %(base_dir))
    corpus = dict()
    for folder in corpus_folders:
        corpus_name = os.path.basename(folder)
        meta = glob("%s/*meta*" %(folder))
        if len(meta)>0:
            corpus[corpus_name] = meta
    return corpus


def get_corpus(analysis_dir):
    corpus_folders = find_directories("%s/corpus" %(analysis_dir))
    corpus = dict()
    for folder in corpus_folders:
        corpus_name = os.path.basename(folder)
        sentences = glob("%s/*sentences*" %(folder))
        if len(sentences)>0:
            corpus[corpus_name] = sentences
    return corpus


def subset_corpus(corpus,delim="_",position=0):
    '''subset_corpus
    return a subset of the corpus based on 
    splitting the file name by delimiter "delim" 
    [default is "_"] at position "position" [default is 0]
    '''
    subset = dict()
    for document in corpus:
        topic = os.path.basename(document).split(delim)[position]
        if topic in subset:
            subset[topic].append(document)
        else:
            subset[topic] = [document]
    return subset






'''
xml: general functions for parsing xml

'''

import re
import os
import tarfile
import xmltodict
from wordfish.utils import get_url

def get_xml_tree(tarball):
    '''get_xml_tree:
    get xmltree from a tarball (.tar.gz) file
    (appropriate for pubmed open content)
    Parameters
    ==========
    tarball: path (str)
        full path to .tar.gz file
    Returns
    =======
    raw: str
        raw text from the xml file
    '''
    if re.search("[.tar.gz]",tarball):
        raw = extract_xml_compressed(paper)
    else:
        raw = read_xml(tarball)
    return raw


def recursive_text_extract(xmltree,element_name):
    '''recursive_text_extract
    Return text for xml tree elements with element_name
    Parameters
    ==========
    xmltree: an xmltree object

    '''
    text = []
    queue = []
    record_ids = []
    for elem in reversed(list(xmltree)):
        queue.append(elem)

    while (len(queue) > 0):
        current = queue.pop()
        if current.text != None:
            text.append(current.text)
        if element_name in current.keys():
            record_ids.append(current.text)
        if len(list(current)) > 0:
            for elem in reversed(list(current)):
                queue.append(elem)

    return record_ids


def extract_xml_compressed(tarball):
    '''extract_xml_compressed
    Read XML from compressed file
    '''
    tar = tarfile.open(tarball, 'r:gz')
    for tar_info in tar:
        if os.path.splitext(tar_info.name)[1] == ".nxml":
            print "Extracting text from %s" %(tar_info.name)
            file_object = tar.extractfile(tar_info)
            return file_object.read().replace('\n', '')


def read_xml_url(url):
    page = get_url(url)
    return xmltodict.parse(page)
    

def read_xml(xml):
    '''read_xml
    Extract text from xml or nxml file
    Parameters
    ==========
    xml: path/str
        path to xml file
    Returns
    =======
    text with newlines replaced with ""
    '''
    with open (xml, "r") as myfile:
        return myfile.read().replace('\n', '')


def crop_text(text,remove_before="<abstract>",remove_after="<ref-list>"):
    '''crop_text
    Cut out article sections we aren't interested in
    '''
    # Remove everything before abstract
    start = re.compile(remove_before)
    end = re.compile(remove_after)
    start = start.search(text)
    end = end.search(text)
    return text[start.start():end.start()]

def remove_formatting(text,to_remove=None):
    if to_remove == None:
        to_remove = ["<italic>","<bold>","<p>","<sub>","<table>","<td>","<tr>"]
    for remove in to_remove:
        text = text.replace(remove,"")
        text = text.replace(remove.replace("<","</"),"")
    return text

def search_text(text,terms):
    '''search_text
    Search text for list of terms, return list of match counts
    '''
    vector = np.zeros(len(terms))
    for t in range(0,len(terms)):
        expression = re.compile("\s%s\s|\s%s\." %(terms[t],terms[t]))
        match = expression.findall(text)
        vector[t] = len(match)
    return vector



'''
analysis.py

part of the wordfish python package: extracting relationships of terms from corpus

'''

from wordfish.nlp import text2sentences, sentence2words, find_phrases
from wordfish.utils import read_json
from numpy import average
from glob import glob
import numpy
import gensim
import pandas
import os

# Training ######################################################################
class TrainSentences(object):
    def __init__(self, text_files):
       self.files = text_files
    def __iter__(self):
        for input_file in self.files:
            for text in file(input_file, "rb"):
                for line in text2sentences(text):            
                    words = sentence2words(line)
                    if len(words) < 3: continue    
                    yield words

def train_word2vec_model(text_files):
    sentences = TrainSentences(text_files)
    model = gensim.models.Word2Vec(sentences, size=300, workers=8, min_count=40)
    return model

def train_lda_model(text_files):
    sentences = TrainSentences(text_files)
    model = gensim.models.Word2Vec(sentences, size=300, workers=8, min_count=40)
    return model


# Classification ###############################################################

class DeepTextAnalyzer(object):
    def __init__(self, word2vec_model):
        # https://dato.com/learn/gallery/notebooks/deep_text_learning.html
        """
        Construct a DeepTextAnalyzer using the input Word2Vec model
        :param word2vec_model: a trained Word2Vec model
        """
        self.model = word2vec_model
    def text2vectors(self,text):
        """
        Convert input text into an iterator that returns the corresponding vector representation of each
        word in the text, if it exists in the Word2Vec model
        :param txt: input text
        :param is_html: if True, then extract the text from the input HTML
        :return: iterator of vectors created from the words in the text using the Word2Vec model.
        """
        words = sentence2words(text)
        words = [w for w in words if w in self.model]
        if len(words) != 0:
            for w in words:
                yield self.model[w]
    def text2mean_vector(self,text,read_file=True):
        """
        Calculate the average vector representation of the input text
        :param txt: input text
        :param is_html: is the text is a HTML
        :return the average vector of the vector representations of the words in the text  
        """
        if read_file == True:
            text = open(text,"rb").read().strip("\n")
        vectors = self.text2vectors(text)
        vectors_sum = next(vectors, None)
        if vectors_sum is None:
            return None
        count = 1.0
        for v in vectors:
            count += 1
            vectors_sum = numpy.add(vectors_sum,v)
        # calculate the average vector and replace +infy and -inf with numeric values 
        avg_vector = numpy.nan_to_num(vectors_sum/count)
        return avg_vector


def save_models(models,base_dir):
    '''
    save_models: should be a dictionary with tags as keys, models as value
    '''
    for model_key,model in models.iteritems():
        model.save("%s/analysis/models/%s.word2vec" %(base_dir,model_key))

def build_models(corpus,model_type="word2vec"):
    models = dict()
    print "Training models..."
    for corpus_id,sentences in corpus.iteritems():
        try:
            if model_type == "word2vec":
                models[corpus_id] = train_word2vec_model(sentences)
            else:
                models[corpus_id] = train_lda_model(sentences)
        except:
            print "Error building model for %s" %(corpus_id)
            pass
    return models


def load_models(base_dir,model_keys=None):
    if isinstance(model_keys,str): model_keys = [model_keys]
    models = dict()
    model_dir = "%s/analysis/models" %(base_dir)
    if model_keys == None:
        model_keys = glob("%s/*.word2vec" %(model_dir))
        model_keys = [os.path.basename(x).replace(".word2vec","") for x in model_keys]
    for model_key in model_keys:
        model_file = "%s/%s.word2vec" %(model_dir,model_key)
        if os.path.exists(model_file):
            models[model_key] = gensim.models.Word2Vec.load(model_file)
    return models

def export_vectors(models,output_dir,sep="\t"):
    # Export vectors
    for model_name,model in models.iteritems():
        print "Processing %s" %(model_name)
        vecs = extract_vectors(model)
        vecs.to_csv("%s/%s.tsv" %(output_dir,model_name),sep=sep)


def extract_vectors(model,vocab=None):
    if vocab==None:
        vocab = model.vocab.keys()
    length = len(model.__getitem__(vocab[0]))
    vectors = pandas.DataFrame(columns=range(length))
    for v in vocab:
        vectors.loc[v] = model.__getitem__(v)
    return vectors

# This is very slow - likely we can extract matrix from model itself
def extract_similarity_matrix(model,vocab=None):
    if vocab==None:
        vocab = model.vocab.keys()
    nvocab = len(model.vocab.keys())
    simmat = pandas.DataFrame(columns=vocab)
    for v in range(len(vocab)):
        term = vocab[v]
        print "parsing %s of %s: %s" %(v,len(vocab),term)
        sims = model.most_similar(term,topn=nvocab)
        values = [sims[x][1] for x in range(len(sims)) if sims[x][0] in vocab]
        labels = [sims[x][0] for x in range(len(sims)) if sims[x][0] in vocab]
        simmat.loc[v+1,labels] = values
    return simmat

def export_models_tsv(models,base_dir,vocabs=None):
    '''
    export_models_tsv: 
    models: dict
        should be a dictionary with tags as keys, models as value
    vocabs: list
        the vocabulary to extract from each model, must be
        same length as model
    '''
    if vocabs != None:
        if len(vocabs)!=len(models):
            print "There must be a vocab specified for each model."
            return
    count=0
    for tag,model in models.iteritems():
        if vocabs==None:
            export_model_tsv(model,tag,base_dir,vocabs)
        else:
            export_model_tsv(model,tag,base_dir,vocabs[count])
            count+=1
            

def export_model_tsv(model,tag,base_dir,vocab=None):
    '''
    export_model_tsv: 
    model: gensim.Word2Vec object
    tag: tag corresponding to model name (corpus)
    '''
    df = extract_similarity_matrix(model,vocab=vocab)
    df.to_csv("%s/analysis/models/%s.tsv" %(base_dir,tag),sep="\t")
    return df


def vocab_term_intersect(terms,model):
    '''
    Finds the intersection of a terms data structure and model.
    Uses a stemming method to stem both, and find overlap.
    Returns a dictionary, with 

       [key]--> model name
       [value]--> tuples in form
       (term_index,vocab_index,term,vocab)

    repeat: the number of times to go over the text (appropriate for
    longer text with possible repetition of terms)
    '''
    vocab = model.vocab.keys()
    intersects = dict()
    for tag,term_set in terms.iteritems():
        if isinstance(term_set,dict):
            names = [x["name"] for x in term_set["nodes"].values()] 
            phrases = find_phrases(words=names,vocabulary=vocab)
            # Returns (word index,vocab index, word, vocab)
            # (697, 3160, u'somatosensation', 'somatosensory')
            intersects[tag] = phrases
    return intersects


def get_labels(meta):
    labels = []
    for r in range(len(meta)):
        meta_file = meta[r]
        labels = numpy.unique(labels + read_json(meta_file)["labels"]).tolist() 
    return labels


def featurize_to_corpus(model,meta,size=300,fillna=True):
    '''featurize_to_corpus
    generate average feature vectors for a set of documents and their labels based
    on an existing model (model). The meta json file should describe text and labels
    '''   
    analyzer = DeepTextAnalyzer(model)
    vectors = pandas.DataFrame(columns=range(size))
    # Get all unique term names from the meta objects
    term_names = get_labels(meta=meta)
    # Create a matrix of labels, keep track of which abstracts are labeled
    labels = pandas.DataFrame(columns=term_names)
    for r in range(len(meta)):
        meta_file = meta[r]
        text = meta_file.replace("_meta","_sentences")
        label = os.path.basename(text).split("_")[0]
    # Build a model for everyone else
    if label not in vectors.index:
        try:
            print "Processing %s of %s" %(r,len(meta))
            vectors.loc[label] = analyzer.text2mean_vector(text)
            labels.loc[label,read_json(meta_file)["labels"]] = 1
        except:
            pass
    count = 1
    for r in range(len(meta)):
        meta_file = meta[r]
        post = meta_file.replace("_meta","_sentences")
        # Build a model by taking the mean vector
        if count not in vectors.index:
            try:
                print "Processing %s of %s" %(r,len(meta))
                vectors.loc[count] = analyzer.text2mean_vector(post)
                labels.loc[count,read_json(meta_file)["labels"]] = 1
                count+=1
            except:
                pass
    if fillna:
        labels = labels.fillna(0)
        vectors = vectors.fillna(0)
    return vectors,labels



'''

browser:
 
part of the wordfish python package: extracting relationships of terms from corpus
this set of functions is for parsing content in a web browser

'''

from selenium import webdriver
from selenium.webdriver.common.keys import Keys

def get_browser():
    return webdriver.Firefox()

def get_page(browser,url):
    browser.get(url)

# Run javascript and get output
def run_javascript(browser,code):
    return browser.execute_script(code)

# Blogger Functions
# A function to click to the next page
def next_page(browser,page_number):
    browser.execute_script('''
        var page_number = parseInt(arguments[0])
        document.getElementsByClassName("gsc-cursor-page")[page_number].click()
        ''',page_number)

# A function to get blog urls on a page
def get_blogs(browser):
    return browser.execute_script('''
        var results = document.getElementsByClassName('gsc-url-top')
        var urls = [];
        for (i = 0; i < results.length-2; i++) { 
            urls.push(results[i].children[0].textContent)
        }
        return urls''')



from wordfish.vm import generate_database_url, custom_app_download, generate_app
from wordfish.plugin import get_plugins, get_corpus, get_terms
from flask import Flask, render_template, request, send_file
from wordfish.utils import copy_directory, make_zip
from werkzeug import secure_filename
import webbrowser
import tempfile
import shutil
import random
import os

# WORDFISH SERVER CONFIGURATIOn #####################################
class WordfishServer(Flask):

    def __init__(self, *args, **kwargs):
            super(WordfishServer, self).__init__(*args, **kwargs)

            # download repo on start of application
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)

app = WordfishServer(__name__)

"""
Get value from a form field

"""
def get_field(request,fields,value):
    if value in request.form.values():
        fields[value] = request.form[value]
    return fields

"""
Main function for starting interface to generate an application
"""
# Step 0: User is presented with base interface
@app.route('/')
def start():
    #TODO: we could define custom analysis object here 

    return render_template('index.html')

# STEP 1: Validation of user input for app
@app.route('/validate',methods=['POST'])
def validate():
    logo = None
    if request.method == 'POST':
        fields = dict()
        for field,value in request.form.iteritems():
            fields[field] = value
             
        # Prepare temp folder with cloned wordfish-python and wordfish-plugins
        custom_app_download(tmpdir=app.tmpdir)
    
        # Get valid plugins to present to user
        valid_plugins = get_plugins("%s/plugins/" %(app.tmpdir),load=True)
        corpus = get_corpus(valid_plugins)
        terms = get_terms(valid_plugins)
 
        return render_template('plugins.html',
                                corpus=str(corpus),
                                terms=str(terms),
                                this_many_corpus = len(corpus),
                                this_many_terms = len(terms))

    return render_template('index.html')

# STEP 2: User must select plugins to include
@app.route('/select',methods=['POST'])
def select():
    if request.method == 'POST':
        fields = dict()
        for field,value in request.form.iteritems():
            fields[field] = value

        # Retrieve experiment folders 
        valid_plugins = get_plugins("%s/plugins" %(app.tmpdir),load=True)
        plugins =  [x[0]["tag"] for x in valid_plugins]
        selected_plugins = [x for x in fields.values() if x in plugins]
        plugin_folders = ["%s/plugins/%s" %(app.tmpdir,x) for x in selected_plugins]

        # Add to the application
        app_dest = "%s/app" %(app.tmpdir)
        generate_app(app_dest,app_repo="%s/python"%app.tmpdir,
                              plugin_repo="%s/plugins"%app.tmpdir,
                              plugins=selected_plugins)

        # Clean up
        clean_up("%s/plugins"%(app.tmpdir))

        # Zip up application
        zipped = make_zip(app_dest,"%s/wordfish.zip" %(os.getcwd()))    

        return send_file(zipped, attachment_filename='wordfish.zip', as_attachment=True)
    return render_template('index.html')


def clean_up(dirpath):
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)

# Visualization functions
@app.route('/sims')
def sims():
    # For now, assume files are in running directory
    cwd = os.getcwd()
    input_files = glob("%s/*.tsv" %(cwd))
    return render_template('word2vec.html',tsv_matrices=input_files)

# This is how the command line version will run
def main(port=8088):
    if port==None:
        port=8088
    print "Let's Fish!"
    webbrowser.open("http://localhost:%s" %(port))
    app.run(host="0.0.0.0",debug=True,port=port)
    
if __name__ == '__main__':
    app.debug = True
    app.run(host='0.0.0.0')



'''
owl
data structure for deepdive python
THESE FUNCTIONS ARE NOT WRITTEN / TESTED YET

Owl:   Protege output (owl files)
       ._read_ontology:   Read in methods from ontology (owl) file
       .get_ontology:     Return list of classes
       ._parse_ontology:  Returns list of classes with alternate terms

'''


# Owl------------------------------------------------------------------------------
class Owl:
    def __init__(self,infile):
        self.file = infile  # name of the owl input file
        self.rawmethod = self._read_ontology()
        self.methods = self._parse_ontology()

    ''' Return raw list of classes'''
    def get_classes():
      return self.methods

    ''' Read methods from owl file - no spaces'''
    def _read_ontology(self):
      methods = []
      filey = open(self.file,"r")
      filey = filey.readlines()
      islabel = re.compile("^<rdfs:label>*")
      # The method names will not have http
      for line in filey:
        line = line.replace(" ","").replace("\n","")
        if islabel.match(line):
          method = line.replace("<rdfs:label>","").replace(" ","").replace("</rdfs:label>","")
          methods.append(method)
      return methods

    '''Return dictionary of methods with alternate terms'''
    def _parse_ontology(self):
      methods = []
      # TODO: we should add the "searchable term" as a field in ontology?

      # Add spaces to methods - return in list
      # TODO: return in dict with alternate terms
      for m in self.rawmethod:
        # add a space to capital letters
        capitals = [c for c, (letter) in enumerate(m) if letter.isupper()]
        tmp = []
        for c in capitals:
          tmp.append(m[:c] + " ")
        tmp.append(m[c:len(m)])
        tmp = ''.join(tmp).strip().lower()
        methods.append(tmp)
      
      # TODO: Debug why the first methods in the list are single letters
      # For now get rid of them!
      methods = [m for m in methods if len(m) > 1]
      return methods
      # TODO: Use JMort method to look up alternative terms for each




'''
vm: part of the wordfish python package
functions for working with virtual machines, repos

'''
from wordfish.utils import copy_directory, get_template, save_template, get_installdir, sub_template
from wordfish.plugin import get_plugins, move_plugins, load_plugins, go_fish, load_plugin
from git import Repo
from glob import glob
import tempfile
import shutil
import numpy
import re
import os


def generate_app(app_dest,app_repo=None,plugin_repo=None,plugins=None):
    '''generate 
    create an app (python module) from a template and list of plugins
    Parameters
    ==========
    app_dest: path
        folder to generate the app (python module) in
    app_repo: path
        full path to wordfish-python (template) repo
    plugins: list
        full paths to plugin folders for inclusion
    '''
    # We can only generate an in a folder that does not exist, to be safe
    if not os.path.exists(app_dest):
        if app_repo == None or plugin_repo == None:
            tmpdir = custom_app_download()
            if app_repo == None:
                app_repo = "%s/python" %(tmpdir)     
            if plugin_repo == None:
                plugin_repo = "%s/plugins" %(tmpdir)     

        # Copy app skeleton to destination
        copy_directory(app_repo,app_dest)
        valid_plugins = get_plugins(plugin_repo)

        # If the user wants to select a subset
        if plugins != None:
            subset_plugins = [x for x in valid_plugins if os.path.basename(x) in plugins]
            valid_plugins = subset_plugins  

        # Generate the requirements.txt from template to include all python dependencies for plugins
        generate_requirements(valid_plugins,app_dest)

        # Copy valid plugins into app_repo
        move_plugins(valid_plugins,app_dest)

        # Generate run commands
        setup_extraction(valid_plugins,app_dest)

    else:
        print "Folder exists at %s, cannot generate." %(battery_dest)


def custom_app_download(tmpdir=None,repo_types=["plugins","python"]):
    '''custom_app_download
    download both repos for the code and plugins to validate
    '''
    acceptable_types = ["plugins","python"]
    if not tmpdir:
        tmpdir = tempfile.mkdtemp()
    if isinstance(repo_types,str):
        repo_types = [repo_types]
    for repo in repo_types:
        if repo in acceptable_types:
            download_repo("https://www.github.com/word-fish/wordfish-%s" %(repo),"%s/%s" %(tmpdir,repo))
        else:
            print "%s is not an acceptable option for repo_types." %(repo)
    return tmpdir


def download_repo(repo_url,tmpdir=None):
    '''download_repo
    Download a github repo to a destination
    Parameters
    ==========
    tmpdir: path
       full path to download repo to
    repo_url: path 
       full url path to repo
    destination: path
       the full path to the destination for the repo
    '''
    if not tmpdir:
        tmpdir = tempfile.mkdtemp()
    Repo.clone_from(repo_url,tmpdir)
    return tmpdir


def generate_database_url(dbtype=None,username=None,password=None,host=None,table=None,template=None):
    '''
    Generate a database url from input parameters, or get a template
        dbtype: the type of database, must be one of "mysql" or "postgresql"
        username: username to connect to the database
        password: password to connect to the database
        host: database host
        table: table in the database
        template: can be one of "mysql" "sqlite3" or "postgresql" If specified, all other parameters are
                  ignored, and a default database URL is produced to work in a Vagrant VM
    
    '''
    if not template:
        if not dbtype or not username or not password or not host or not table:
            print "Please provide all inputs: dbtype,username,password,host,table."
        else:
            return "%s://%s:%s@%s/%s"  %(dbtype,
                                         username,
                                         password,
                                         host,
                                         table)
    elif template == "mysql":
        return "mysql://wordfish:wordfish@localhost:3306/wordfish"
    elif template == "postgresql":
        return "postgresql://wordfish:wordfish@localhost:5432/wordfish"


def generate_requirements(valid_plugins,app_dest):
    '''
    generate_requirements will generate a custom setup.py from a template
    this will include the python dependencies for valid plugins
    '''
    plugin_folder = os.path.dirname(valid_plugins[0])
    v = load_plugins(valid_plugins)

    # Add plugin dependencies on other plugins, even if user has not selected
    plugin_deps = [x[0]["dependencies"]["plugins"] for x in v if "plugins" in x[0]["dependencies"]]
    for plugin_dep in plugin_deps:
        for p in plugin_dep:
            if "%s/%s" %(plugin_folder,p) not in valid_plugins:
                valid_plugins.append("%s/%s" %(plugin_folder,p))

    # Get list of python dependencies for setup.py
    python_deps = [x[0]["dependencies"]["python"] for x in v if "python" in x[0]["dependencies"]]
    dependencies = []
    for python_dep in python_deps:
        dependencies = dependencies + python_dep
    dependencies = numpy.unique(dependencies).tolist()

    # Get requirements.txt to use as a template
    requirements = get_template("%s/requirements.txt" %(app_dest)).split("\n")
    new_requirements = numpy.unique(requirements + dependencies).tolist()
    new_requirements = [x for x in new_requirements if len(x) > 0]

    # Save new requirements
    save_template("%s/requirements.txt" %(app_dest),"\n".join(new_requirements))


def setup_extraction(valid_plugins,app_dest):
    '''
    setup_extractions will generate a run time script
    directory in the application scripts folder, and then, for
    each plugin for which some field is True,
    write a line to a job script (that can be run with launch or slurm)
    to call the function.

    '''
    script_directory = "%s/wordfish/scripts" %app_dest
    if not os.path.exists(script_directory):
        os.mkdir(script_directory)
    extraction_script = "%s/run_first.job" %(script_directory)
    for valid_plugin in valid_plugins:
        plugin = load_plugin(valid_plugin)[0]
        if plugin["corpus"] == "True" or plugin["terms"] == "True":
            go_fish(plugin["tag"],extraction_script)

def init_scripts(scripts_dir,output_base):
    '''init_scripts:
    move job running scripts from template into user
    scripts directory. Template substitutions will be done at
    this step, to ensure that calling the init_scripts function
    from a different folder will produce scripts to run with correct
    analysis directory path
    '''
    installdir = get_installdir()
    scripts_to_move = glob("%s/scripts/*" %(installdir))
    for script in scripts_to_move:
        script_template = get_template(script)
        script_name = os.path.basename(script)
        script_copy = "%s/scripts/%s" %(output_base,script_name)       
        save_template(script_copy,script_template)              
        

def make_plugin_folders(analysis_dir):
    '''
    Makes output directories for all plugins defined in the package
    for which a corpus, terms, or relationships extraction is defined.
    '''
    installdir = get_installdir()
    installed_plugins = get_plugins("%s/plugins" %(installdir),load=True)
    folders = ["corpus","terms","relations"]
    for installed_plugin in installed_plugins:
        tag = installed_plugin[0]["tag"]
        for folder in folders:
            make_plugin_folder(analysis_dir,folder,tag,installed_plugin[0][folder],"True")

def make_plugin_folder(analysis_dir,folder,tag,field_name,field_value):
    if field_name == field_value:
        plugin_folder = "%s/%s/%s" %(analysis_dir,folder,tag)
        if not os.path.exists(plugin_folder):
            os.mkdir(plugin_folder)



# Create a matrix of labels, keep track of which abstracts are labeled
labels = pandas.DataFrame(columns=term_names)

for r in range(len(meta)):
    meta_file = meta[r]
    text = meta_file.replace("_meta","_sentences")
    label = os.path.basename(text).split("_")[0]
    # Build a model for everyone else
    if label not in vectors.index:
        try:
            print "Processing %s of %s" %(r,len(meta))
            vectors.loc[label] = analyzer.text2mean_vector(text)
            labels.loc[label,read_json(meta_file)["labels"]] = 1
        except:
            pass




'''
plugin.py

part of the wordfish python package: extracting relationships of terms from corpus
this set of functions finds and validates plugins for inclusion in the package

TODO: These functions will be written when it is time to create the web app portion of this.

'''
from wordfish.utils import find_directories, remove_unicode_dict, copy_directory, add_lines, wordfish_home
from glob import glob
import inspect
import json
import os
import nltk
import numpy
import pandas

def get_validation_fields():
    """
    Returns a list of tuples (each a field)
      required for a valid json
      (field,value,type)
          field: the field name
          value: indicates minimum required entires
          type: indicates the variable type

    """
    return [("name",1,str),
            ("tag",1,list),
            ("corpus",1,str), 
            ("terms",1,str),
            ("dependencies",0,dict), 
            ("arguments",0,str),
            ("contributors",0,str), 
            ("doi",0,str)]
           
def notvalid(reason):
    print reason
    return False


def validate(plugin_folder):
    """
    validate:
    takes a plugin folder, and looks for validation based on:

    - config.json
    - in future this will be more complex, for now just validating the config!

    """

    meta = load_plugin(plugin_folder)

    # Don't return the template folder as an option
    if meta[0]["tag"] == "template":
        return False

    if meta != None:
        if len(meta)>1:
            return notvalid("config.json has length > 1, not valid.")
        fields = get_validation_fields()
        for field,value,ftype in fields:
            if value != 0:
                # Field must exist in the keys
                if field not in meta[0].keys():
                    return notvalid("config.json is missing field %s" %(field))
                if meta[0][field] == "":
                    return notvalid("config.json must be defined for field %s" %(field))
    else:
        return notvalid("cannot load config.json")
    return True   

def get_corpus(plugins,return_field=None):
    '''get_corpus:
    return list of corpus for plugins that have it defined
    Parameters
    ==========
    plugins: dict
        list of plugins to look through (dict)
    return_file: str
        The name of a field to return, if the entire
        data structure is not desired
    Returns
    =======
    corpus: list
       list of plugins with corpus
    '''
    corpus = filter_by_field(plugins,"corpus","True")
    if return_field != None:
        corpus = [x[0][return_field] for x in corpus]
        corpus = numpy.unique(corpus).tolist()
    return corpus
    
def filter_by_field(plugins,field,field_value):
    passing = []
    for plugin in plugins:
        if plugin[0][field] == field_value:
            passing.append(plugin)
    return passing

def get_terms(plugins,return_field=None):
    '''get_terms:
    return list of corpus for plugins that have it defined
    Parameters
    ==========
    plugins: dict
        list of plugins to look through
    Returns
    =======
    terms: list
       list of plugins with terms defined
    '''
    terms = filter_by_field(plugins,"terms","True")
    if return_field != None:
        terms = [x[0][return_field] for x in terms]
        terms = numpy.unique(terms).tolist()
    return terms


def get_plugins(plugin_repo=None,load=False):
    '''get_plugins from a downloaded wordfish-plugins folder
    download plugin repo to destination folder
    Parameters
    ==========
    plugin_repo: path to plugins repo
    load: boolean
        if True, returns loaded json (dict). If false,
        returns paths to json files
    '''
    if plugin_repo == None:
        tmpdir = custom_app_download(repo_types=["plugins"])
        plugin_repo = "%s/plugins" %(tmpdir)

    plugins = find_directories(plugin_repo)
    valid_plugins = [p for p in plugins if validate(p)]
    print "Found %s valid plugins" %(len(valid_plugins))
    if load == True:
        valid_plugins = load_plugins(valid_plugins)
    return valid_plugins


def load_plugins(plugin_folders):
    '''load_plugins:
    read in the config.json from plugin folders, return json object
    Parameters
    ==========
    plugin_folders: list
        a list of plugin folders from the wordfish-plugins repo
    Returns
    =======
    plugins:
        list of json objects, one for each plugin
    '''
    if isinstance(plugin_folders,str):
        plugin_folders = [plugin_folders]
    plugins = []
    for plugin_folder in plugin_folders:
        plugins.append(load_plugin(plugin_folder))
    return plugins

def load_plugin(plugin_folder):
    """

    load_plugin:
    reads in the config.json for an:
         
        plugin folder: full path to a plugin folder

    """
    fullpath = os.path.abspath(plugin_folder)
    config = "%s/config.json" %(fullpath)
    if not os.path.exists(config):
        return notvalid("config.json could not be found in %s" %(plugin_folder))
    try: 
        meta = json.load(open(config,"r"))
        meta = remove_unicode_dict(meta[0])
        return [meta]
    except ValueError as e:
        print "Problem reading config.json, %s" %(e)


def move_plugins(valid_plugins,app_dest):
    '''
    Moves valid plugins into the app folder for install
    
        valid_plugins: a list of full paths to valid plugin folders
        app_dest: full path to app destination folder
    '''

    moved_plugins = []
    for valid_plugin in valid_plugins:
        try:
            plugin_folder = os.path.basename(valid_plugin)
            copy_directory(valid_plugin,"%s/wordfish/plugins/%s" %(app_dest,plugin_folder))
            moved_plugins.append(valid_plugin)
        except:
           print "Cannot move %s, will not be installed." %(valid_plugin)
    return moved_plugins

# EXTRACTION JOBS ############################################################

def go_fish(tag,extraction_script):
    line_to_add = "python -c 'from wordfish.plugins.%s.functions import go_fish; go_fish()'" %(tag)
    add_lines(script=extraction_script,lines_to_add=[line_to_add])

def generate_job(func,category,inputs=None,batch_num=1):
    '''
    generate_job
    Parameters
    ==========
    func: str
        name of function to call in plugin functions.py
    category: str
        must be one of "terms" or "corpus" or "relations" corresponding to output folder
    inputs: dict
        key should be arg name, and value should be list of string args as input to func
        If inputs are not specified, it is assumed that the function will be called once
        with no inputs.
    batch_num: int
        the number of jobs to package into one job. For example, batch_num=100 will run
        func with 100 of the input items specified. Each is still written to its own
        output file.
    '''
    # Get name of calling plugin
    home = wordfish_home()
    cf = inspect.currentframe()    
    caller = inspect.getouterframes(cf, 2)
    tag = os.path.dirname(caller[1][1]).split("/")[-1]
    script = "wordfish.plugins.%s.functions" %(tag)
    output_dir = ' output_dir="%s/%s/%s"' %(home,category,tag) 

    # script name to add jobs to
    extraction_script = "%s/scripts/run_extractions_%s.job" %(home,tag)

    lines_to_add = []      
    if category in ["corpus","terms","relations"]:
        if inputs == None:
            lines_to_add.append("python -c 'from %s import %s; %s(%s)'" %(script,func,func,output_dir))
        else:
            formatted_inputs = ""
            # First collect all string args - this means same for all scripts
            for varname,elements in inputs.iteritems():
                if isinstance(elements,str):
                    single_input = format_single_input(varname,elements)
                    formatted_inputs = "%s%s" %(formatted_inputs,single_input)
                          
            # Now collect lists, must be equal length
            input_lists = dict()
            for varname,elements in inputs.iteritems():
                if isinstance(elements,list):
                    if len(input_lists)>0:
                        if len(input_lists.values()[0]) == len(elements):    
                            input_lists[varname] = elements
                    else:
                        input_lists[varname] = elements

            # If we have no input lists, just write the job with single args
            if len(input_lists) == 0:
                formatted_inputs = formatted_inputs.strip(",")
                lines_to_add.append("python -c 'from %s import %s; %s(%s,%s)'" %(script,func,func,output_dir,formatted_inputs))
            else:
                N = len(input_lists.values()[0])
                iters = int(numpy.ceil(N/float(batch_num)))
                start = 0
                for i in range(1,iters+1):
                    formatted_instance = formatted_inputs
                    if i==N:
                        end = N
                    else:
                        end = i*batch_num
                    for varname,elements in input_lists.iteritems():
                        new_input = format_inputs(varname,elements[start:end])
                        formatted_instance = "%s%s" %(formatted_instance,new_input)
                    start = end
                    formatted_instance.strip(",")
                    lines_to_add.append("python -c 'from %s import %s; %s(%s,%s)'" %(script,func,func,output_dir,formatted_instance))

        # Add lines
        add_lines(script=extraction_script,lines_to_add=lines_to_add)
    

def format_single_input(varname,element):
    if isinstance(element,str):
        element = '"%s"' %(element)
    return " %s=%s," %(varname,element) 

def format_inputs(varname,elements):
    elements = ['"%s"' %(x) if isinstance(x,str) else x for x in elements]
    return " %s=[%s]" %(varname,",".join([str(x) for x in elements])) 



'''
nlp:part of the wordfish python package: extracting relationships of terms from corpus

functions for simple natural language processing

'''
from textblob import TextBlob, Word
from nltk.corpus import stopwords
from nltk.stem.porter import *
from nltk.stem import *
import nltk.data
import numpy
import pandas
import gensim
import re

def remove_nonenglish_chars(text):
    return re.sub("[^a-zA-Z]", " ", text)
    
def text2sentences(text,remove_non_english_chars=True):
    tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')    
    if remove_non_english_chars:
        text = remove_nonenglish_chars(text)
    for s in tokenizer.tokenize(text):
        yield s

def processText(text):
    '''combines text2sentences and sentence2words'''
    vector = []
    for line in text2sentences(text):            
        words = sentence2words(line)
        vector = vector + words
    return vector

def sentence2words(sentence,remove_stop_words=True,lower=True):
    if isinstance(sentence,list): sentence = sentence[0]
    re_white_space = re.compile("\s+")
    stop_words = set(stopwords.words("english"))
    if lower: sentence = sentence.lower()
    words = re_white_space.split(sentence.strip())
    if remove_stop_words:
        words = [w for w in words if w not in stop_words]
    return words

def do_stem(words,return_unique=True,remove_non_english_words=True):
    '''do_stem
    Parameters
    ==========    
    words: str/list
        one or more words to be stemmed
    return_unique: boolean (default True)
        return unique terms
    '''
    stemmer = PorterStemmer()
    if isinstance(words,str):
        words = [words]
    stems = []
    for word in words:
        if remove_non_english_words:
            word = re.sub("[^a-zA-Z]", " ", word)
        stems.append(stemmer.stem(word))
    if return_unique:
        return numpy.unique([s.lower() for s in stems]).tolist()
    else:
        return stems

def get_total_words(text):
    '''get_total_words:
    get total words in a text (dict, string, or list)
    Parameters
    ==========
    text: str,dict,list
        some text content to parse to count total words
    Returns
    =======
    totalwords: int
        total count of words
    '''
    totalwords = 0

    # Dictionary
    if isinstance(text,dict):
        for label,sentences in text.iteritems():
            if isinstance(sentences,str):
                sentences = [sentences]
            for sentence in sentences:
                blob =  TextBlob(sentence)
                words = do_stem(blob.words)
                totalwords += len(words)
        return totalwords    

    # String or list
    elif isinstance(text,str):
        text = [text]
    for sentence in text:
        blob =  TextBlob(sentence)
        words = do_stem(blob.words)
        totalwords += len(words)
    return totalwords


def get_term_counts(terms,text):
    '''get_term_counts:
    a wrapper for get_term_counts_dict and get_term_counts_list
    will return a pandas data frame of counts for a list of terms of interst
    Parameters
    ==========
    text: dict,list,str
        some text content to parse to count a number of terms
    terms: str,list
        one or more terms to be stemmed and counted in the text
    Returns
    =======
    totalwords: int
        total count of words
    '''
    if isinstance(text,dict):
        return get_term_counts_dict(terms,text)
    elif isinstance(text,str):
        text = [text]
    elif isinstance(text,list):
        return get_term_counts_list(terms,text)


def get_term_counts_list(terms,text):
    # Convert words into stems
    stems = do_stem(terms)

    # data frame hold counts
    counts = pandas.DataFrame(0,columns=["count"],index=stems)

    for sentence in text:
        blob =  TextBlob(sentence)
        words = do_stem(blob.words)
        words = [w for w in words if w in stems]
        counts.loc[words] = counts.loc[words] + 1
    return counts        
    

def get_term_counts_dict(terms,text):
    # Convert words into stems
    stems = do_stem(terms)

    # data frame hold counts
    counts = pandas.DataFrame(0,columns=["count"],index=stems)

    for label,sentences in text.iteritems():
        if isinstance(sentences,str):
            sentences = [sentences]
        for sentence in sentences:
            blob =  TextBlob(sentence)
            words = do_stem(blob.words)
            words = [w for w in words if w in stems]
            counts.loc[words] = counts.loc[words] + 1
    return counts

# Return list of stemmed phrases
def stem_phrases(phrases):
    stemmed = []
    for phrase in phrases:
        phrase = phrase.split(" ")
        if isinstance(phrase,str):
            phrase = [phrase]
        single_stemmed = do_stem(phrase)
        stemmed.append(" ".join(single_stemmed).encode("utf-8"))
    return stemmed

def get_match(phrasematch,entirephrase,found_indices):
    '''
    get_match:

    Function to get a match: start, length, text, from a sentence

    Returns dictionary with:
  
    start_index
    length
    text
    found_indices: updated binary [0,1] list of found indices in entirephrase

    '''

    full_concept = phrasematch.split(" ")
    foundmatch = True
    # We should not find words that have already been found :)
    findices = [i for i in range(0,len(found_indices)) if found_indices[i] == 1]
    for found_index in findices:
        entirephrase[found_index] = "XXXXXXXXXXXXXXXX"
    indices = []
    for word in full_concept:
        if word in entirephrase:
            indices.append(entirephrase.index(word))
        # Missing any one word, not a match
        else:
            foundmatch = False
    if len(numpy.unique(indices)) == len(full_concept):
        for i in range(0,len(indices)-1):
            # Not in chronological order +1, not a match
            if indices[i]+1 != indices[i+1]:
                foundmatch=False
    # Missing any one word, not a match
    else:
        foundmatch = False
    if foundmatch == True:
        start_index = entirephrase.index(full_concept[0])
        length = len(full_concept)
        text = entirephrase[start_index:start_index+length]   
        # Update found indices
        found_indices[start_index:start_index+length]=1   
    else:
        start_index = 0
        length = 0
        text = ""
    result = {"start_index":start_index,
              "length":length,
              "text":text,
              "found_indices":found_indices}
    return result

def find_phrases(words,vocabulary,repeat=1):
    '''
    words: a list of words
    vocabulary: a list of words / phrases to find in the words
    repeat: the number of times to run over the phrase
    (in case of repeats of same in one sentence)
    
    Returns:
    
    (words_index,vocab_index,word,vocab)
    
    '''
    vocabulary = numpy.unique(vocabulary).tolist()
    vocabulary = [v.encode("utf-8") for v in vocabulary]
    # We will stem phrases, and search for them across the stemmed words
    vocab_stemmed = stem_phrases(vocabulary)
    stemmed = [s.encode("utf-8") for s in do_stem(words,return_unique=False)]
    # Make a long regular expression
    regexp = "*|".join(vocab_stemmed) + "*"
    phrases = []

    # Make lookups to return to original vocab and terms
    vocab_lookup = make_lookup(vocabulary,vocab_stemmed)
    words_lookup = make_lookup(words,stemmed)

    # We run it twice in case of repeats in a sentence
    for r in range(0,repeat):
        # Search the sentence for any concepts:
        if re.search(regexp," ".join(stemmed)):
            for c in range(0,len(stemmed)):
                for v in range(len(vocab_stemmed)):
                    single_stemmed = vocab_stemmed[v]
                    if re.match("%s" %(stemmed[c]),single_stemmed):
                        phrases.append((c, v, words_lookup[stemmed[c]],vocab_lookup[vocab_stemmed[v]]))            
    return phrases

def make_lookup(original_list,new_list):
    lookup = dict()
    for x in range(len(new_list)):
        lookup[new_list[x]] = original_list[x]
    return lookup




'''
terms.py

part of the wordfish python package: extracting relations between terms from corpus

this set of functions works with different plugins (in plugins folder) to produce input terminologies
to search for in corpus

'''
from wordfish.utils import save_pretty_json, find_directories, read_json
from uuid import uuid4
from glob import glob
import pandas
import nltk
import os

def download_nltk():
    '''download_nltk
    download nltk to home
    '''
    home=os.environ["HOME"]
    download_dir = "%s/nltk_data" %home
    print "Downloading nltk to %s" %(download_dir)
    if not os.path.exists(download_dir):
        import nltk
        nltk.download('all')
    return "%s/nltk_data" %(home)


def save_relations(relations,output_dir=None):
    '''save_relationships
    Parameters
    =========
     output_dir: path
        path to save output. If none, will just return json dictionary
    relations: list of tuples [(source,target,relation)]
        if defined, all keys must be in input_terms. Not yet decided what a "relation" should be, but for now assume you can have it be a string or number.
    Returns
    =======
         list of links
         [{"source":"node1","target":"node2","value":0.5}]

    '''
    # Not sure why anyone would do this, but might as well check
    links = []

    # Save relations
    for tup in relations:
        pair = [tup[0],tup[1]]
        pair.sort()
        if output_dir is not None:
            output_file = "%s/%s_relations.json" %(output_dir,"_".join(pair).replace(" ",""))
            relation = {"source":tup[0],"target":tup[1],"value":tup[2]}
            links.append(relation)
            if not os.path.exists(output_file):
                tmp = save_pretty_json(relation,output_file)

    return links            

def save_terms(input_terms,output_dir=None):
    '''save_terms
    Parameters
    =========
    input_terms: list,dict
        a list or dictionary of terms. if meta data are used to describe the  input_terms, provide the input_terms as a dictionary with a dictionary to define {"meta_label":"meta_value"}
     output_dir: path
        path to save output. If none, will just return json dictionary
    Returns
    =======
    links: dict
        dictionary structure with the following format (parallel to what many d3
        algorithms use to define graphs)

        {"nodes":[{"name":"node1"},
                 {"name":"node2"}],
         }

    '''
    nodes = []
    ids = []
    if isinstance(input_terms,str):
        input_terms = [input_terms]
    if isinstance(input_terms,list):
        input_terms = [x.lower() for x in input_terms]
        for t in range(len(input_terms)):
            term = input_terms[t]
            nodes.append({"name":term.lower(),"uid":str(t)})
            ids.append(term.lower())
    elif isinstance(input_terms,dict):
        for node, meta in input_terms.iteritems():
            meta["uid"] = str(node).lower()
            nodes.append(meta)
            ids.append(str(node).lower())
    else:
        print "Invalid input_terms, must be str, dict, or list."
        return

    result = {"nodes":nodes}
    if output_dir is not None:
        tmp = save_pretty_json(result,"%s/terms.json" %(output_dir))
    return result


def get_terms(analysis_dir,subset=True):
    '''
    For all terms defined, and relationships for the terms, parse into a single data structure
    This (maybe) won't work for larger datasets (we will use a database) but it will for testing.

        nodes:

            {"[plugin]::[uid]":[node]}

    Parameters
    ==========
    analysis_dir: path
        full path to analysis directory
    subset: boolean
        if True, returns terms in dictionary based on source tag. Default==False    
    '''

    nodes = dict()
    edges = dict()

    terms_dir = "%s/terms" %(os.path.abspath(analysis_dir))
    if os.path.exists(terms_dir):
        term_plugins = find_directories(terms_dir)


        nodes = dict()
        edges = dict()
        results = dict()

        for term_plugin in term_plugins:
            plugin_name = os.path.basename(term_plugin)

            if subset:
                nodes = dict()
                edges = dict()

            # Here we parse together terms
            if os.path.exists("%s/terms.json" %term_plugin):
                terms_json = read_json("%s/terms.json" %term_plugin)["nodes"]
                for node in terms_json:
                    if "uid" in node:
                        uid = "%s::%s" %(plugin_name,node["uid"])
                    else:
                        feature_name = node["name"].replace(" ","_")
                        uid = "%s::%s" %(plugin_name,feature_name) 
                    nodes[uid] = node

            # Here we parse together relationships
            # Currently only supported for terms within the same family
            if os.path.exists("%s/term_relationships.json" %term_plugin):
                terms_json = read_json("%s/term_relationships.json" %term_plugin)["edges"]
                for relation in terms_json:
                    uid_1 = "%s::%s" %(plugin_name,relation["source"])
                    uid_2 = "%s::%s" %(plugin_name,relation["target"])
                    relation_uid = "%s<>%s" %(uid_1,uid_2)
                    edges[relation_uid] = {"source": uid_1,
                                           "target": uid_2,
                                           "value": relation["value"]}

            result = {"nodes":nodes,"edges":edges}
            if subset:
                results[plugin_name] = result
    
    if subset:
        result = results
    else:
        result = {"all":result}
    # Return the result to user with all edges and nodes defined
    if analysis_dir is not None:
        tmp = save_pretty_json(result,"%s/terms/terms.json" %(analysis_dir))
    return result

def get_relations(base_dir,tags=None,read=False):
    edges = dict()
    if isinstance(tags,str):
        tags = [tags]
    relations_dir = "%s/relations" %(os.path.abspath(base_dir))
    if tags == None:
        tags = [os.path.basename(x) for x in find_directories(relations_dir)]
    for tag in tags:
        print "Finding relations for %s" %(tag)
        relations_files = glob("%s/%s/*_relations.json" %(relations_dir,tag))
        if len(relations_files) != 0:
            if read:
                edges[tag] = read_relations(relations_files)
            else:
                edges[tag] = relations_files       
    return edges

# This is inefficient, we really need a document database
def get_relations_df(base_dir,tags=None):
    if isinstance(tags,str):
        tags = [tags]
    relations_dir = "%s/relations" %(os.path.abspath(base_dir))
    if tags == None:
        tags = [os.path.basename(x) for x in find_directories(relations_dir)]
    for tag in tags:
        print "Finding relations for %s" %(tag)
        relations_files = glob("%s/%s/*_relations.json" %(relations_dir,tag))
        term_names = numpy.unique([x.split("_")[0] for x in relations_files]).tolist()
        edges = pandas.DataFrame(columns=term_names,index=term_names)
        for r in range(len(relations_files)):
            relation_file = relations_files[r]
            print "Parsing %s of %s" %(r,len(relations_files))
            term1,term2=os.path.basename(relation_file).split("_")[0:2]      
            edges.loc[term1,term2] = read_json(relation_file)["value"]
            edges.loc[term2,term1] = read_json(relation_file)["value"]
        relations[tag] = edges
    return relations

def read_relations(relations_list,search_expression=None):
    if search_expression != None:
        expression = re.compile(search_expression)
        return [read_json(x) for x in relations_files if expression.search(x)]
    else:
        relations = []
        for x in range(len(relations_files)):
            print "Parsing %s of %s" %(x,len(relations_files))  
            relations.append(read_json(relations_files[x])) 




'''
run_analysis.py

USAGE:

python run_analysis.py

Your wordfish project home directory should be defined as WORDFISH_HOME

This script will merge all terms and corpus into a common framework, and then produce simple
analyses related to term frequency in corpus, etc. Custom analyses scripts can be based off of this. 
For now, the method is as follows:

    - Merge terms and relationships into a common corpus
    - For all text extract features with deep learning (word2vec)
    - Build classifiers to predict labels from vectors

'''

# First train simple word2vec model with different corpus
from wordfish.analysis import build_models, save_models, export_models_tsv, load_models, extract_similarity_matrix, export_vectors, featurize_to_corpus
from wordfish.models import build_svm
from wordfish.corpus import get_corpus, get_meta, subset_corpus
from wordfish.terms import get_terms
from wordfish.utils import mkdir
import os

base_dir = os.environ["WORDFISH_HOME"]

# Setup analysis output directories
analysis_dir = mkdir("%s/analysis" %(base_dir))
model_dir = mkdir("%s/models" %(analysis_dir))
vector_dir = mkdir("%s/vectors" %(analysis_dir))

# Generate more specific corpus by way of file naming scheme
corpus = get_corpus(base_dir)
reddit = corpus["reddit"]
disorders = subset_corpus(reddit)
corpus.update(disorders)


# Train corpus specific models
models = build_models(corpus)

# Export models to tsv, export vectors, and save
save_models(models,base_dir)
export_models_tsv(models,base_dir)
export_vectors(models,output_dir=vector_dir)


### Working with models #######################################################################

# Find the most similar words for a query word:
model = load_models(base_dir)["neurosynth"]
model.most_similar("anxiety")
# [('aggression', 0.77308839559555054), 
#   ('stress', 0.74644440412521362), 
#   ('personality', 0.73549789190292358), 
#   ('excessive', 0.73344630002975464), 
#   ('anhedonia', 0.73305755853652954), 
#   ('rumination', 0.71992391347885132), 
#   ('distress', 0.7141801118850708), 
#   ('aggressive', 0.7049030065536499), 
#   ('craving', 0.70202392339706421), 
#   ('trait', 0.69775849580764771)]

# Corpus context is important
model = load_models(base_dir)["reddit"]
model.most_similar("anxiety")
# [('crippling', 0.64760375022888184), 
# ('agoraphobia', 0.63730186223983765), 
# ('generalized', 0.61023455858230591), 
# ('gad', 0.59278655052185059), 
# ('hypervigilance', 0.57659250497817993), 
# ('bouts', 0.56644737720489502), 
# ('depression', 0.55617612600326538), 
# ('ibs', 0.54766887426376343), 
# ('irritability', 0.53977066278457642), 
# ('ocd', 0.51580017805099487)]

# perform addition and subtraction with vectors
model.most_similar(positive=['anxiety',"food"])
# [('ibs', 0.50205761194229126), 
# ('undereating', 0.50146859884262085), 
# ('boredom', 0.49470821022987366), 
# ('overeating', 0.48451068997383118), 
# ('foods', 0.47561675310134888), 
# ('cravings', 0.47019645571708679), 
# ('appetite', 0.46869537234306335), 
# ('bingeing', 0.45969703793525696), 
# ('binges', 0.44506731629371643), 
# ('crippling', 0.4397256076335907)]

model.most_similar(positive=['bipolar'], negative=['manic'])
# [('nos', 0.36669495701789856), 
# ('adhd', 0.36485755443572998), 
# ('autism', 0.36115738749504089), 
# ('spd', 0.34954413771629333), 
# ('cptsd', 0.34814098477363586), 
# ('asperger', 0.34269329905509949), ('schizotypal', 0.34181860089302063), ('pi', 0.33561226725578308), ('qualified', 0.33355745673179626), ('diagnoses', 0.32854354381561279)]

model.similarity("anxiety","depression")
#0.67751728687122414

model.doesnt_match(["child","autism","depression","rett","cdd","young"])
#'depression'

# Get the raw vector for a word
model["depression"]

# Extract a pairwise similarity matrix
wordfish_sims = extract_similarity_matrix(models["neurosynth"])



# Classication ############################################################################
# Predict neurosynth abstract labels with pubmed neurosynth corpus

# Load model and meta for neurosynth, meaning labels for each text
model = load_models(base_dir,"neurosynth")["neurosynth"]
meta = get_meta(base_dir)["neurosynth"]
vectors,labels = featurize_to_corpus(model,meta)
classifiers = build_svm(vectors=vectors,labels=labels,kernel="linear")



from setuptools import setup, find_packages
import codecs
import os


setup(
    # Application name:
    name="wordfish",

    # Version number (initial):
    version="0.7",

    # Application author details:
    author="Poldracklab",
    author_email="vsochat@stanford.edu",

    # Packages
    packages=find_packages(),

    # Data files
    include_package_data=True,
    zip_safe=False,

    # Details
    url="http://www.github.com/word-fish/wordfish-python",

    license="LICENSE.txt",
    description="Infrastructure for finding relationships between terms in corpus of interest.",
    keywords='wordfish, nlp, text parsing',

    install_requires = ["numpy","nltk","gitdb","smmap","flask","gitpython","gensim","graphlab-create","xmltodict"],

    entry_points = {
        'console_scripts': [
            'wordfish=wordfish.interface:main',
        ],
    },

)



'''
run_slurm.py

A basic script to read in a list of commands, parse into slurm job scripts,
and run the jobs. This will be improved! For now, modify the bottom of the 
script to change the default submission args.

'''

import os
import sys
script_dir = os.path.dirname(os.path.realpath(__file__))
job_dir = "%s/.job" %(script_dir)
out_dir = "%s/.out" %(script_dir)

### MODIFY SUBMISSION ARGS HERE ###########################################
queue_name = "normal"
time = "2-00:00"
memory = 64000
##############


if len(sys.argv) < 2:
    print 'Please provide job file with list of single commands as input.'

else:
    jobfile = sys.argv[1]

    if not os.path.exists(job_dir):
        os.mkdir(job_dir)
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    # Read in the jobfile, and write a job for each line
    filey = open(jobfile,"rb")
    commands = filey.readlines()
    filey.close()

    # Name the job based on input script
    name = os.path.basename(jobfile).split(".")[0]

    for c in range(len(commands)):
        command = commands[c]
        jobfile = open("%s/%s_%s.job" %(job_dir,name,c),'w')
        jobfile.writelines("#!/bin/bash\n")
        jobfile.writelines("#SBATCH --job-name=%s_%s.job\n" %(name,c))
        jobfile.writelines("#SBATCH --output=%s/%s_%s.out\n" %(out_dir,name,c))
        jobfile.writelines("#SBATCH --error=%s/%s_%s.err\n" %(out_dir,name,c)) 
        jobfile.writelines("#SBATCH --time=%s\n" %(time)) 
        jobfile.writelines("#SBATCH --mem=%s\n" %(memory))   
        jobfile.writelines(command)  
        jobfile.close()
        os.system('sbatch -p %s -n 1 %s/%s_%s.job' %(queue_name,job_dir,name,c))



'''
models.py

part of the wordfish python package: specific machine learning models

'''

from numpy.random import choice
from sklearn import svm
import numpy
import pandas
import sys
import os

def build_svm(vectors,labels,kernel="linear",positive_label=1,negative_label=0,testing_holdout=0.2):
    '''build_svm
    vectors: a data frame of vectors to be used for training and testing
    labels: a pandas data frame of objects (rows) and labels (columns)
    labels.index should correspond with vectors.index
    '''
    classifiers = pandas.DataFrame(columns=["accuracy","N","term","N_train","N_test","testNPos",
                               "testNNeg","trainNPos","trainNNeg","TP","FP","TN","FN","kernel"])
    term_names = labels.columns.tolist()
    for term in term_names:
        if "%s_%s" %(term,kernel) not in classifiers.index:
            data_Y = labels[term]
            positive_examples = data_Y.index[data_Y==positive_label]
            negative_examples = data_Y.index[data_Y==negative_label]
            # hold out some percentage of each for testing
            testSizePositive = int(numpy.ceil(len(positive_examples)*testing_holdout))
            testSizeNegative = int(numpy.ceil(len(negative_examples)*testing_holdout))
            testN = testSizePositive + testSizeNegative
            testPos = choice(positive_examples,testSizePositive).tolist()
            testNeg = choice(negative_examples,testSizeNegative).tolist()
            test = vectors.loc[testPos+testNeg]
            test_Y = [term if x in testPos else "not %s" %(term) for x in test.index]
            trainPos = [x for x in positive_examples if x not in testPos]
            trainNeg = [x for x in negative_examples if x not in testNeg]
            trainN = len(trainPos) + len(trainNeg)
            train = vectors.loc[trainPos+trainNeg]
            train_Y = [term if x in trainPos else "not %s" %(term) for x in train.index]
            clf = svm.SVC(kernel=kernel)
            clf.fit(train, train_Y)
            # Make a single prediction based on vectors
            predictions=[]
            for t in test.index:
                predictions.append(clf.predict(test.loc[t])[0])
            correct = float(len([x for x in range(len(predictions)) if predictions[x]==test_Y[x]]))
            accuracy = correct / len(test_Y)
            positive_predictions = [x for x in range(len(predictions)) if predictions[x] == term]
            negative_predictions = [x for x in range(len(predictions)) if predictions[x] == "not %s" %term]
            positive_actual = [x for x in range(len(test_Y)) if test_Y[x] == term]
            negative_actual = [x for x in range(len(test_Y)) if test_Y[x] == "not %s" %term]
            TP = len([x for x in positive_actual if x in positive_predictions])
            TN = len([x for x in negative_actual if x in negative_predictions])
            FP = len([x for x in negative_actual if x in positive_predictions])
            FN = len([x for x in positive_actual if x in negative_predictions])
            classifiers.loc["%s_%s" %(term,kernel)] = [accuracy,len(data_Y),term,trainN,testN,len(testPos),len(testNeg),len(trainPos),len(trainNeg),TP,FP,TN,FN,kernel]

    return classifiers
